package proxy;

import io.netty.bootstrap.ServerBootstrap;
import io.netty.buffer.ByteBuf;
import io.netty.buffer.ByteBufOutputStream;
import io.netty.buffer.Unpooled;
import io.netty.channel.Channel;
import io.netty.channel.ChannelException;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelHandler;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.ChannelOption;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.epoll.Epoll;
import io.netty.channel.epoll.EpollEventLoopGroup;
import io.netty.channel.epoll.EpollServerSocketChannel;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.ServerSocketChannel;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import io.netty.handler.codec.base64.Base64;
import io.netty.handler.timeout.ReadTimeoutHandler;
import io.netty.util.concurrent.Future;
import io.netty.util.concurrent.GenericFutureListener;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.net.IDN;
import java.net.InetAddress;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Queue;
import java.util.UUID;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executors;
import java.util.concurrent.FutureTask;

import javax.imageio.ImageIO;

import com.google.common.base.Charsets;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Queues;
import com.google.common.util.concurrent.Futures;
import com.google.common.util.concurrent.ListenableFuture;
import com.google.common.util.concurrent.ListenableFutureTask;
import com.google.common.util.concurrent.ThreadFactoryBuilder;

import proxy.network.EnumPacketDirection;
import proxy.network.MessageDeserializer;
import proxy.network.MessageDeserializer2;
import proxy.network.MessageSerializer;
import proxy.network.MessageSerializer2;
import proxy.network.NetHandlerHandshake;
import proxy.network.NetHandlerPlayServer;
import proxy.network.NetworkManager;
import proxy.network.ThreadQuickExitException;
import proxy.packet.play.server.S40PacketDisconnect;
import proxy.util.ChatColor;
import proxy.util.ChatComponentText;
import proxy.util.Profile;
import proxy.util.ServerInfo;
import proxy.util.User;
import proxy.util.VersionId;
import proxy.util.LazyLoadBase;
import proxy.util.Log;
import proxy.util.PlayerInfo;

public class Proxy {
	public static final LazyLoadBase<NioEventLoopGroup> NIO_EVENT_LOOP = new LazyLoadBase<NioEventLoopGroup>() {
		protected NioEventLoopGroup load() {
			return new NioEventLoopGroup(0, (new ThreadFactoryBuilder()).setNameFormat("Netty Server IO #%d").setDaemon(true).build());
		}
	};
	public static final LazyLoadBase<EpollEventLoopGroup> EPOLL_EVENT_LOOP = new LazyLoadBase<EpollEventLoopGroup>() {
		protected EpollEventLoopGroup load() {
			return new EpollEventLoopGroup(0, (new ThreadFactoryBuilder()).setNameFormat("Netty Epoll Server IO #%d").setDaemon(true).build());
		}
	};

	private volatile boolean isAlive;
	private volatile boolean running = true;
	
	private final List<ChannelFuture> endpoints = Collections.<ChannelFuture>synchronizedList(Lists.newArrayList());
	private final List<NetworkManager> networkManagers = Collections.<NetworkManager>synchronizedList(Lists.newArrayList());
	private final Queue<FutureTask<?>> futureTaskQueue = Queues.<FutureTask<?>>newArrayDeque();
	private final Map<String, User> users = Maps.newHashMap();
	private final Map<String, NetHandlerPlayServer> players = Maps.newHashMap();
	private final Thread serverThread;
	
	private int compression = -1;
	private boolean epoll = false;
	private boolean register = true;
	private boolean checkCase = true;
	private String forwardHost = "lian.dd";
	private int forwardPort = 25565;
	private String proxyHost = "";
	private int proxyPort = 25566;
	private int minPassLength = 8;
	private int maxAttempts = 5;
	private ServerInfo status = getStatus(new File("icon.png"), ChatColor.AQUA + "Server\n" + ChatColor.GREEN + "Test!!", 90, 10, "Test1",
			"Test2 ...", "Test #3 !!");

	private static void addFaviconToStatusResponse(ServerInfo info, File file) {
		if(file.isFile()) {
			ByteBuf buf = Unpooled.buffer();
			try {
				BufferedImage img = ImageIO.read(file);
				if(img.getWidth() != 64 || img.getHeight() != 64)
					throw new IllegalArgumentException("Icon must be 64x64 pixels");
				ImageIO.write(img, "PNG", new ByteBufOutputStream(buf));
				ByteBuf base64 = Base64.encode(buf);
				info.setIcon("data:image/png;base64," + base64.toString(Charsets.UTF_8));
			}
			catch(Exception e) {
				Log.error(e, "Couldn't load server icon");
			}
			finally {
				buf.release();
			}
		}
	}

	public static ServerInfo getStatus(File icon, String motd, int max, int current, String... lines) {
		ServerInfo info = new ServerInfo();
		info.setMotd(new ChatComponentText(motd));
		info.setVersion(new VersionId("1.8.9", 47));
		addFaviconToStatusResponse(info, icon);
		info.setInfo(new PlayerInfo(max, current));
		Profile[] profiles = new Profile[Math.min(lines.length, 12)];
		for(int z = 0; z < profiles.length; z++) {
			profiles[z] = new Profile(Proxy.getOfflineUUID(lines[z]), lines[z]);
		}
		info.getInfo().setList(profiles);
		return info;
	}

	public static UUID getOfflineUUID(String name) {
		return UUID.nameUUIDFromBytes(("OfflinePlayer:" + name).getBytes(Charsets.UTF_8));
	}

	public static void main(String[] args) {
		Thread.currentThread().setName("Proxy thread");
		Proxy proxy = new Proxy();
		proxy.run();
	}

	public Proxy() {
		this.serverThread = Thread.currentThread();
		this.isAlive = true;
	}
	
	public void run() {
		Log.info("Starting login proxy on %s:%d", this.proxyHost.isEmpty() ? "0.0.0.0" : this.proxyHost, this.proxyPort);
		try {
			this.addLanEndpoint(this.proxyHost.isEmpty() ? null : InetAddress.getByName(IDN.toASCII(this.proxyHost)), this.proxyPort);
		}
		catch(IOException e) {
			Log.error(e, "Could not bind to port");
			return;
		}
		Runtime.getRuntime().addShutdownHook(new Thread(new Runnable() {
			public void run() {
				Proxy.this.terminateEndpoints();
			}
		}, "Proxy shutdown thread"));
		while(this.running) {
	        synchronized (this.futureTaskQueue)
	        {
	            while (!this.futureTaskQueue.isEmpty())
	            {
	            	FutureTask<?> task = this.futureTaskQueue.poll();
					try {
						task.run();
						task.get();
					}
					catch(ExecutionException e1) {
						if(!(e1.getCause() instanceof ThreadQuickExitException))
							Log.error(e1, "Error executing task " + task);
					}
					catch(InterruptedException e2) {
						Log.error(e2, "Error executing task " + task);
					}
	            }
	        }
			this.networkTick();
			try {
				Thread.sleep(50L);
			}
			catch(InterruptedException e) {
			}
		}
		this.terminateEndpoints();
		Log.info("Proxy stopped");
	}

	public void addLanEndpoint(InetAddress address, int port) throws IOException {
		synchronized(this.endpoints) {
			Class<? extends ServerSocketChannel> oclass;
			LazyLoadBase<? extends EventLoopGroup> lazyloadbase;

			if(Epoll.isAvailable() && this.epoll) {
				oclass = EpollServerSocketChannel.class;
				lazyloadbase = EPOLL_EVENT_LOOP;
				Log.info("Using epoll channel type");
			}
			else {
				oclass = NioServerSocketChannel.class;
				lazyloadbase = NIO_EVENT_LOOP;
				Log.info("Using default channel type");
			}

			this.endpoints
					.add(((ServerBootstrap)((ServerBootstrap)(new ServerBootstrap()).channel(oclass)).childHandler(new ChannelInitializer<Channel>() {
						protected void initChannel(Channel p_initChannel_1_) throws Exception {
							try {
								p_initChannel_1_.config().setOption(ChannelOption.TCP_NODELAY, Boolean.valueOf(true));
							}
							catch(ChannelException var3) {
								;
							}

							p_initChannel_1_.pipeline().addLast((String)"timeout", (ChannelHandler)(new ReadTimeoutHandler(30)))
									.addLast((String)"splitter", (ChannelHandler)(new MessageDeserializer2()))
									.addLast((String)"decoder", (ChannelHandler)(new MessageDeserializer(EnumPacketDirection.SERVERBOUND)))
									.addLast((String)"prepender", (ChannelHandler)(new MessageSerializer2()))
									.addLast((String)"encoder", (ChannelHandler)(new MessageSerializer(EnumPacketDirection.CLIENTBOUND)));
							NetworkManager networkmanager = new NetworkManager(EnumPacketDirection.SERVERBOUND);
							Proxy.this.networkManagers.add(networkmanager);
							p_initChannel_1_.pipeline().addLast((String)"packet_handler", (ChannelHandler)networkmanager);
							networkmanager.setNetHandler(new NetHandlerHandshake(Proxy.this, networkmanager));
						}
					}).group((EventLoopGroup)lazyloadbase.getValue()).localAddress(address, port)).bind().syncUninterruptibly());
		}
	}

	public void terminateEndpoints() {
		this.isAlive = false;

		for(ChannelFuture channelfuture : this.endpoints) {
			try {
				channelfuture.channel().close().sync();
			}
			catch(InterruptedException var4) {
				Log.error("Interrupted whilst closing channel");
			}
		}
	}

	public void networkTick() {
		synchronized(this.networkManagers) {
			Iterator<NetworkManager> iterator = this.networkManagers.iterator();

			while(iterator.hasNext()) {
				final NetworkManager networkmanager = (NetworkManager)iterator.next();

				if(!networkmanager.hasNoChannel()) {
					if(!networkmanager.isChannelOpen()) {
						iterator.remove();
						networkmanager.checkDisconnected();
					}
					else {
						try {
							networkmanager.processReceivedPackets();
						}
						catch(Exception exception) {
							Log.error(exception, "Failed to handle packet for " + networkmanager.getRemoteAddress());
							final ChatComponentText chatcomponenttext = new ChatComponentText("Internal server error");
							networkmanager.sendPacket(new S40PacketDisconnect(chatcomponenttext), new GenericFutureListener<Future<? super Void>>() {
								public void operationComplete(Future<? super Void> p_operationComplete_1_) throws Exception {
									networkmanager.closeChannel(chatcomponenttext);
								}
							});
							networkmanager.disableAutoRead();
						}
					}
				}
			}
		}
	}

	private <V> ListenableFuture<V> callFromMainThread(Callable<V> callable) {
		if(!this.isMainThread()) {
			ListenableFutureTask<V> task = ListenableFutureTask.<V>create(callable);

			synchronized(this.futureTaskQueue) {
				this.futureTaskQueue.add(task);
				return task;
			}
		}
		else {
			try {
				return Futures.<V>immediateFuture(callable.call());
			}
			catch(Exception e) {
				return Futures.immediateFailedCheckedFuture(e);
			}
		}
	}

	public void schedule(Runnable task) {
		this.<Object>callFromMainThread(Executors.callable(task));
	}

	public boolean isMainThread() {
		return Thread.currentThread() == this.serverThread;
	}

	public int getCompression() {
		return this.compression;
	}

	public ServerInfo getStatus() {
		return this.status;
	}

	public String getForwardHost() {
		return this.forwardHost;
	}

	public int getForwardPort() {
		return this.forwardPort;
	}

	public boolean isUsingEPoll() {
		return this.epoll;
	}
	
	public boolean canRegister() {
		return this.register;
	}
	
	public boolean isCheckingCase() {
		return this.checkCase;
	}

	public int getMinimumPasswordLength() {
		return this.minPassLength;
	}

	public int getMaximumPasswordAttempts() {
		return this.maxAttempts;
	}
	
	public User getUser(String user) {
		return this.users.get(user.toLowerCase(Locale.US));
	}
	
	public void setUser(String user, User password) {
		this.users.put(user.toLowerCase(Locale.US), password);
	}
	
	public void setLoggedIn(String user, NetHandlerPlayServer handler) {
		this.players.put(user.toLowerCase(Locale.US), handler);
	}
	
	public void setLoggedOut(String user) {
		this.players.remove(user.toLowerCase(Locale.US));
	}
	
	public boolean isLoggedIn(String user) {
		return this.players.containsKey(user.toLowerCase(Locale.US));
	}
}
