package proxy.network;

import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelFutureListener;
import proxy.Proxy;
import proxy.packet.login.client.C00PacketLoginStart;
import proxy.packet.login.client.C01PacketEncryptionResponse;
import proxy.packet.login.server.S00PacketDisconnect;
import proxy.packet.login.server.S02PacketLoginSuccess;
import proxy.packet.login.server.S03PacketEnableCompression;
import proxy.util.ChatComponent;
import proxy.util.ChatComponentText;
import proxy.util.Log;

public class NetHandlerLoginServer implements INetHandler {
	private static enum LoginState {
		INIT, WAITING, DONE;
	}

	private final Proxy proxy;
	private final NetworkManager connection;

	private LoginState state = LoginState.INIT;
	private int timer;
	private String username;
	
	private static boolean isValidUser(String name) {
		for(int z = 0; z < name.length(); z++) {
			char ch = name.charAt(z);
			if((ch < '0' || ch > '9') && (ch < 'a' || ch > 'z') && (ch < 'A' || ch > 'Z') && ch != '_')
				return false;
		}
		return true;
	}

	public NetHandlerLoginServer(Proxy proxy, NetworkManager connection) {
		this.proxy = proxy;
		this.connection = connection;
	}

	public void update(NetworkManager connection) {
		if(this.state == LoginState.WAITING) {
			this.tryAcceptPlayer();
		}

		if(this.timer++ == 600) {
			this.closeConnection("Took too long to log in");
		}
	}

	public void closeConnection(String reason) {
		try {
			Log.info("Disconnecting " + this.getConnectionInfo() + ": " + reason);
			ChatComponentText text = new ChatComponentText(reason);
			this.connection.sendPacket(new S00PacketDisconnect(text));
			this.connection.closeChannel(text);
		}
		catch(Exception e) {
			Log.error(e, "Error whilst disconnecting player");
		}
	}

	public void tryAcceptPlayer() {
		if(this.proxy.isLoggedIn(this.username)) {
			this.closeConnection("That username is already taken");
			return;
		}
		this.state = LoginState.DONE;

		if(this.proxy.getCompression() >= 0) {
			this.connection.sendPacket(new S03PacketEnableCompression(this.proxy.getCompression()), new ChannelFutureListener() {
				public void operationComplete(ChannelFuture p_operationComplete_1_) throws Exception {
					NetHandlerLoginServer.this.connection.setCompressionTreshold(NetHandlerLoginServer.this.proxy.getCompression());
				}
			});
		}

		NetHandlerPlayServer handler = new NetHandlerPlayServer(this.proxy, this.connection, this.username);
		this.connection.setNetHandler(handler);
		this.connection.sendPacket(new S02PacketLoginSuccess(Proxy.getOfflineUUID(this.username), this.username));
		handler.setupPlayer();
	}

	public void onDisconnect(NetworkManager connection, ChatComponent reason) {
		Log.info(this.getConnectionInfo() + " lost connection: " + reason);
	}

	public String getConnectionInfo() {
		return this.username != null ? this.username + " (" + this.connection.getRemoteAddress().toString() + ")"
				: String.valueOf(this.connection.getRemoteAddress());
	}

	public void processLoginStart(C00PacketLoginStart packet) {
		if(this.state != LoginState.INIT)
			throw new IllegalStateException("Unexpected hello packet");
		this.username = packet.getProfile();
		if(this.username.length() < 3 || this.username.length() > 16 || !isValidUser(this.username)) {
			this.closeConnection("Invalid username");
			return;
		}
		this.state = LoginState.WAITING;
	}

	public void processEncryptionResponse(C01PacketEncryptionResponse packet) {
		throw new IllegalStateException("Unexpected key packet");
	}
}
